/*
 * Copyright (C) 2016 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blend.
 *
 * blend is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blend is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blend.  If not, see <http://www.gnu.org/licenses/>.
 */

import { Adapter } from "../core/adapter";
import { Metric } from "../core/metric";
import { Source } from "../core/source";
import { Dimension } from "../core/dimension";
import { Clause } from "../core/clause";
import { Filter, FilterOperator } from "../core/filter";
import { AggregationType, RelationType, DataType } from "../common/types";
import { Operation, Opcode } from "../common/expression";
import { View } from "../core/view";
import { Tsort } from "../util/tsort";

/**
 * Information required to make a join clause.
 * The dimension and views that contain this dimension.
 */
interface DimInfo {
    /** Dimension object. */
    dim: Dimension;
    /** Set of views that contain the dimension. */
    views: View[];
}

/**
 * Several translations of a dimension that will be used in different
 * parts of a query
 */
interface DimTranslation {
    /** Translation with the AS clause. Used on SELECT clause. */
    aliased: string;
    /** Translation without the AS clause. USED on GROUP BY clause. */
    noalias: string;
    /** The alias of the dimension. The quoted name. */
    alias: string;
    /** Expansion of a sub dimension. */
    expanded: boolean;
}

/** Translated view, the query which gets all data and its name */
interface QueryAndName {
    /** Query that returns all the data of a view. */
    query: string;
    /** The view name. */
    name: string;
}

/** Dictonary indexed by view name, that returns the QueryAndName object. */
interface QNMap {
    [key: string]: QueryAndName;
}

/**
 * Two Dictionaries, both indexed with a dimension name.
 * Used to get the views where a dimension is.
 * The dimensions also have the Dimension object inside.
 */
interface DimAndNameMap {
    /** Dictionary indexed by dimension name that returns DimInfo */
    dimensions: {[key: string]: DimInfo};
    /**
     * Dictionary indexed by dimension name that returns
     * a view where this dimension is. Used as right side
     * of a join clause.
     */
    views: {[key: string]: View};
}

/**
 * Generic SQL adapter.
 * Contains most implementation to perform a SQL query.
 * However, which database has its peculiarities, so this
 * adapter must be extended to each SGBD.
 */
export abstract class SQLAdapter extends Adapter {
    /**
     * Translate a view to a SQL query.
     * @param view - View to be translated.
     */
    public getQueryFromView(view: View): string {
        const map = this.buildPartials(view);
        const topSort = Tsort.view(view);
        const partials = topSort.map((i) => {
            return (map[i]) ? map[i] : {query : "", name: i};
        }).filter((i) => {
            return i.query !== "";
        }).map((i) => {
            return i.name + " AS (" + i.query + ")";
        });

        let withClause = "";
        if (partials.length > 0) {
            withClause =  "WITH "  + partials.join(", ") + " ";
        }

        let sort = "";
        if (view.sort.length > 0) {
            // Sorting
            const order = view.sort.map((item) => {
                return "\"" + item.name + "\"";
            }).join(",");

            sort = " ORDER BY " + order;
        }

        const dimensions = view.dimensions.map((item) => {
            return "\"" + item.name + "\"";
        }).join(",");

        const metrics = view.metrics.map((item) => {
            return "\"" + item.name + "\"";
        }).join(",");

        let attributes = "";
        if (dimensions.length > 0) {
            if (metrics.length > 0) {
                attributes = dimensions + "," + metrics;
            }

            else {
                attributes = dimensions;
            }
        }

        else {
            attributes = metrics;
        }

        return withClause + "SELECT " + attributes + " FROM " + view.name + sort + ";";
    }

    /**
     * Translates a view to a SQL sub-query. The main diffence between
     * this method and getQueryFromView is that a sub-query cannot have
     * sorting properties. Also a sub-query is called a partial and
     * partials can only have one metric.
     * @param view - View which the partial will be built.
     */
    private buildPartials(view: View): QNMap {
        let op = view.operation;
        let queue: View[] = op.values.map((i) => i);
        const output: QNMap = {};
        output[view.name] = {
            query: this.operation(op, view),
            name: view.name
        };

        const map: {[key: string]: boolean } = {};

        while (queue.length > 0) {
            const partial = queue.shift();
            if (!map[partial.id]) {
                const query = this.operation(partial.operation, partial);
                if (query !== "") {
                    map[partial.id] = true;
                    if (!output[partial.name]) {
                        output[partial.name] = {
                            query: query,
                            name: partial.name
                        };
                    }
                    queue = queue.concat(partial.operation.values);
                }
            }
        }

        return output;
    }

    /**
     * Constructs a query from a view based in a given operation
     * @param op - Operation used to construct the view.
     * @param view - View to be querified.
     */
    private operation(op: Operation, view: View): string {
        switch (op.opcode) {
            case Opcode.REDUCE:
                return this.buildOperation(view, op.values, false);
            case Opcode.JOIN:
                return this.buildOperation(view, op.values, true);
            default:
                // Is unnecessary make a push function. Push = materialized
                // No need for materialized partials
                return "";
        }
    }

    /**
     * Constructs a query to a expecific view, from a set of given partials.
     * @param view - View to be built.
     * @param partials - Partials (other views) required to built the view.
     * @param isJoin - Used to check if join clauses must be added.
     */
    private buildOperation(view: View, partials: View[], isJoin: boolean): string {
        // Mapping, which views the metrics and dimensions are
        const mapping = this.buildMaps(partials);
        // Projection
        const metrics = view.metrics.map((i) => {
            const sourceView = mapping.views[i.name];
            return this.translateMetric(i, sourceView);
        });
        const dimensions = view.dimensions.map((dimension) => {
            let dim = dimension;
            while (!mapping.views[dim.name]) {
                dim = dim.parent;
            }
            const sourceView = mapping.views[dim.name];
            return this.translateDimension(dimension, dim, sourceView);
        });
        const projElem = dimensions.map((i) => i.aliased).concat(metrics);

        // Grouping
        const grouped = dimensions.map((item) => {
            return (item.expanded) ? item.alias : item.noalias;
        });

        // Selection
        const conds = [];
        const clauses = this.orphanClauses(view, partials);
        for (let i = 0; i < clauses.length; ++i) {
            const trClause = this.translateClause(clauses[i], mapping.views);
            if (trClause) {
                conds.push("(" + trClause + ")");
            }
        }

        // Joinning
        if (isJoin) {
            const dimMap = mapping.dimensions;
            for (let i of Object.keys(dimMap)) {
                if (dimMap[i].views.length > 1) {
                    const views = dimMap[i].views;
                    const dim = dimMap[i].dim;
                    const leftSide = this.buildColumn(dim, views.shift().name);
                    while (views.length > 0) {
                        const rightSide = this.buildColumn(dim, views.shift().name);
                        conds.push("(" + leftSide + "=" + rightSide + ")");
                    }
                }
            }
        }

        // Assembly
        const projection = "SELECT " + projElem.join(", ");
        const sources = "FROM " + partials.map((i) => (i.name)).join(", ");
        const selection = (conds.length > 0) ? " WHERE " + conds.join(" AND ") : "";

        let grouping = "";
        if (grouped.length > 0) {
            grouping = " GROUP BY " + grouped.join(",");
        }

        return projection + sources + selection + grouping;
    }

    /**
     * Constructs and returns the dictionaries which inform
     * in which views a given dimension is.
     * @param views - Set of views to be scored.
     */
    private buildMaps(views: View[]): DimAndNameMap {
        let dimMap: {[key: string]: DimInfo} = {};
        let nameMap: {[key: string]: View} = {};

        for (let i = 0; i < views.length; ++i) {
            const mets = views[i].metrics;
            const dims = views[i].dimensions;
            for (let j = 0; j < mets.length; ++j) {
                if (!nameMap[mets[j].name])  {
                    nameMap[mets[j].name] = views[i];
                }

            }

            for (let j = 0; j < dims.length; ++j) {
                if (!dimMap[dims[j].name])  {
                    dimMap[dims[j].name] = {
                        dim: dims[j],
                        views: [views[i]]
                    };
                    nameMap[dims[j].name] = views[i];
                }

                else {
                    dimMap[dims[j].name].views.push(views[i]);
                }
            }
        }

        return {
            dimensions: dimMap,
            views: nameMap
        };
    }

    /**
     * Returns a set o clauses of the given view that is not fulfilled
     * by any of the partials. In other words, clauses that has not been
     * added to the query yet.
     * @param view - View with all clauses.
     * @param partials - Views with some of the clauses.
     */
    private orphanClauses(view: View, partials: View[]): Clause[] {
        let parentClauses: Clause[] = [];
        for (let i = 0; i < partials.length; ++i) {
            parentClauses = parentClauses.concat(partials[i].clauses);
        }

        // return clauses that does not exist in the partials
        return view.clauses.filter((i) => !parentClauses.some((j) => j.id === i.id));
    }

    /**
     * Parse a BlenDB enumeration of agregation functions to a proper string
     * to be used in the SQL query.
     * @param aggrType - Aggregation function.
     * @param origin - Will be applied in a view with origin flag ?
     */
    private getAggregateFunction(aggrType: AggregationType, origin: boolean): string {
        switch (aggrType) {
            case AggregationType.SUM:
                return "SUM";
            case AggregationType.AVG:
                return "AVG";
            case AggregationType.COUNT:
                return (origin) ? "COUNT" : "SUM";
            case AggregationType.MAX:
                return "MAX";
            case AggregationType.MIN:
                return "MIN";
            default:
                return  "";
        }

    }

    /**
     * Parse a BlenDB enumeration of relationships to a proper string
     * to be used in the SQL query.
     * @param relation - The realtion function that will be used.
     * @param arg - The attribute that is the function argument.
     */
    private translateRelation(relation: RelationType, arg: string): string {
        switch (relation) {
            case RelationType.DAY:
                return this.applyRelation("EXTRACT", ["DAY FROM "], [arg]);
            case RelationType.MONTH:
                return this.applyRelation("EXTRACT", ["MONTH FROM "], [arg]);
            case RelationType.YEAR:
                return this.applyRelation("EXTRACT", ["YEAR FROM "], [arg]);
            default:
                return "";
        }

    }

    /**
     * Generic translate a SQL function to a query
     * @param name - Function name.
     * @param args - Function argument names.
     * @param values - Function argument values.
     */
    private applyRelation(name: string, args: string[], values: string[]): string {
        /*
            This adapter uses the concept of functions in Postgres to
            implement BLENDB sub-dimention relations, this functions
            applys the transformation to build the call of a Postgres
            funtion. Note that this function can be native from postgres,
            like EXTRACT, or even implemented on the database.
            This function is short and only used in the translateRelation
            method however is a bit complex and is possible to be used
            several times, because of that is puted appart to make easyer update
            and avoid problems
            Example
            applyRelation ("EXTRACT", "["DAY FROM"]", ["view_0.date"])
            output: EXTRACT(DAY FROM view_0.date)
        */
        return name + "(" + args.map((item, idx) => item + values[idx]).join(",") + ")";
    }

    /**
     * Add quotes and the proper view name to a attribute in a SQL query.
     * @param item - Attribute to get the name quoted.
     * @param viewName - View name used to build the view name.
     */
    private buildColumn (item: Metric|Dimension, viewName: string): string {
        const quotedName = "\"" + item.name + "\"";
        return viewName + "." + quotedName;
    }

    /**
     * Translate Clause object into a SQL query condition.
     * @param clause - Clause to be translated.
     * @param map - Dictionary indexed bu attributes name
     * containing its locations.
     */
    private translateClause(clause: Clause, map: {[key: string]: View}): string {
        const r = clause.filters.map((item) => {
            return this.translateFilter(item, map);
        }).filter((item) => {
            return item !== "";
        });
        return r.join(" OR ");
    }

    /**
     * Translate Filter object into a SQL query condition.
     * @param filter - Filter to be translated.
     * @param map - Dictionary indexed bu attributes name
     * containing its locations.
     */
    private translateFilter(filter: Filter, map: {[key: string]: View}): string {
        if (!map[filter.target.name]) {
            return "";
        }

        const viewName = map[filter.target.name].name;
        const leftSide = this.buildColumn(filter.target, viewName);
        const quotedValue = "'" + filter.value + "'";
        const castedValue = this.typeCast(quotedValue, filter.target.dataType);
        return this.applyOperator(leftSide, castedValue, filter.operator);
    }

    /**
     * Translate Metric object into a SQL query attribute.
     * @param metric - Metric to be translated.
     * @param view - View which contains the metric. Used to create the name
     */
    private translateMetric(metric: Metric, view: View): string {
        const func = this.getAggregateFunction(metric.aggregation, view.origin);
        const quotedName = "\"" + metric.name + "\"";
        const extMetric = func + "(" + this.buildColumn(metric, view.name) + ")";
        return extMetric + " AS " + quotedName;
    }

    /**
     * Translate Dimension object into a SQL query attribute.
     * @param dimension - Dimension to be translated.
     * @param ancestor - Dimension ancestor found in the view
     * @param view - View which contains the dimenion/ancestor.
     *  Used to create the name.
     */
    private translateDimension(dimension: Dimension,
                               ancestor: Dimension,
                               view: View): DimTranslation {
        const quotedName = "\"" + dimension.name + "\"";
        let extDimension = this.buildColumn(ancestor, view.name);
        let aux = dimension;
        let expanded = false;
        while (aux.name !== ancestor.name) {
            extDimension = this.translateRelation(aux.relation, extDimension);
            aux = aux.parent;
            expanded = true;
        }
        return {
            aliased: extDimension + " AS " + quotedName,
            noalias: extDimension,
            alias: quotedName,
            expanded: expanded
        };
    }

    /**
     * Generates a generic insertion query for one register.
     * @param source - Data insertion "location".
     * @param data - Data to be inserted.
     */
    public getQueryFromSource(source: Source, data: any[]): string {
        let consult: string;
        let colums: any[] = [];
        let values: string[] = [];
        let tam = Object.keys(data).length;
        colums = source.fields.map(o => o.name);
        for (let i = 0; i < tam; i++){
            values[i] = data[colums[i]];
        }
        consult = "INSERT INTO " + source.name + " (\"";
        consult = consult.concat(colums.join("\",\""));
        consult = consult.concat("\") VALUES ('");
        consult = consult.concat(values.join("' , '"));
        consult = consult.concat("')");

        return consult;
    }

    /**
     * Parse a filter operation. Varies in each database.
     * @param lSide - Operation left side operator.
     * @param rSide - Operation right side operator.
     * @param op - Operation to be performed.
     */
    protected abstract applyOperator(leftSide: string, rightSide: string, op: FilterOperator): string;
    /**
     * Cast BlenDB types to proper database types.
     * @param quotedValue - SQL query attribute wrapped by quotes.
     * @param dt - Attribute data type.
     */
    protected abstract typeCast(quotedValue: string, dt: DataType): string;
}
