/*
 * Copyright (C) 2019 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

import { configure, getLogger } from "log4js";
import * as fs from "fs";
/**
 * Define the levels that BlenDB log must have
 */
interface LogInterface {
    debug(msg: string, ...additionalInfo: any[]): void;
    info(msg: string, ...additionalInfo: any[]): void;
    warn(msg: string, ...additionalInfo: any[]): void;
    error(msg: string, ...additionalInfo: any[]): void;
}
/**
 * Logging for BlenDB using the framework log4js,
 * using debug, info, warn and error as levels.
 * Hierarchy of levels for showing messages are:
 * debug > info > warn > error
 */
export class Log implements LogInterface {

    private levels = ["debug" , "info" , "warn" , "error"];
    private file: string;
    private level: string;
    private logger: any;
    /**
     * Definitions to use log4js, such as file destination,
     *  level of log and configure log4js.
     */
    constructor(){
        this.file = (process.env["BLENDB_LOG_FILE"]) ? process.env["BLENDB_LOG_FILE"] : "stderr";
        this.level = (process.env["BLENDB_LOG_LEVEL"]) ? process.env["BLENDB_LOG_LEVEL"] : "error";

        let find = this.levels.some( element => {
            return element === this.level;
        });
        if (!find) {
            this.level = "error";
        }

        if (!(this.file === "stderr")) {
            try {
                fs.appendFileSync(this.file, "");
            }
            catch (e) {
                this.file = "stderr";
            }
        }

        /**
         * The reason for the duplication of configure,
         * is because the filename flag from appenders it start
         * the empty file if doesn't exist, even if it's not
         * called from getLogger. Then it won't create an empty
         * file called stderr.
         */
        if (this.file === "stderr") {
            configure({
                appenders: {    defaultLog: { type : "stderr" }
                    },
                categories: {
                              default: { appenders: ["defaultLog"], level: this.level },
                            }
            });
            this.logger = getLogger("defaultLog");
        }else{
            configure({
                appenders: {    blendb: { type: "file", layout: {
                                type: "pattern",
                                pattern: "%d [%p] host: [%h] user: [%x{user}] %n %[%m%n%]",
                                tokens: {
                                    user: function(logEvent) {
                                      return process.env.USER;
                                    }
                                }
                        },
                        filename: this.file }
                    },
                categories: {
                              default: { appenders: ["blendb"], level: this.level }
                            }
            });
            this.logger = getLogger("blendb");
        }
        this.logger.level = this.level;
    }
    /**
     * Get log level
     */
    public getLogLevel(){
        return this.level;
    }
    /**
     * Get lof filename
     */
    public getLogFile(){
        return this.file;
    }
    /**
     * Log level debug
     * @param msg description of log
     * @param additionalInfo useful additional information
     */
    public debug(msg: string, ...additionalInfo: any[]){
        this.emtigLogMessage("debug", msg, additionalInfo);
    }
    /**
     * Log level info
     * @param msg description of log
     * @param additionalInfo useful additional information
     */
    public info(msg: string, ...additionalInfo: any[]){
        this.emtigLogMessage("info", msg, additionalInfo);
    }
    /**
     * Log level warning
     * @param msg description of log
     * @param additionalInfo useful additional information
     */
    public warn(msg: string, ...additionalInfo: any[]){
        this.emtigLogMessage("warn", msg, additionalInfo);
    }
    /**
     * Log level error
     * @param msg description of log
     * @param additionalInfo useful additional information
     */
    public error(msg: string, ...additionalInfo: any[]){
        this.emtigLogMessage("error", msg, additionalInfo);
    }
    /**
     * Handle data between log class and log4js framework
     * @param msgType level of log being use
     * @param msg description of log
     * @param additionalInfo useful additional information
     */
    private emtigLogMessage(logLevel: "debug" | "info" | "warn" | "error" , msg: string, additionalInfo: any[]){
        if (additionalInfo.length){
            this.logger[logLevel](msg, additionalInfo);
        }else{
            this.logger[logLevel](msg);
        }
    }
}
